from typing import Optional, Dict
from abc import ABC, abstractmethod
import numpy as np
import torch

from annadca.dataset import annaDataset
from annadca.io import _save_model, _load_model
        
        
class annaRBM(ABC):
    """Abstract class for the annotation-assisted Restricted Boltzmann Machine (annaRBM)"""

    def __init__(
        self,
        params: Dict[str, torch.Tensor] | None = None,
        device: Optional[torch.device] = torch.device("cpu"),
        dtype: Optional[torch.dtype] = torch.float32,
    ):
        if params is not None:
            self.params = {k : params[k].to(device=device, dtype=dtype) for k in params.keys()}
        else:
            self.params = None

        self.device = device
        self.dtype = dtype


    @torch.jit.export
    def to(
        self,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        """Moves the parameters to the specified device and/or dtype.

        Args:
            device (Optional[torch.device], optional): Device. Defaults to None.
            dtype (Optional[torch.dtype], optional): Dtype. Defaults to None.

        Returns:
            annaRBM: annaRBM instance with the parameters moved to the specified device and/or dtype.
        """
        if device is not None:
            if self.params is not None:
                self.params = {key: self.params[key].to(device) for key in self.params}
            self.device = device

        if dtype is not None:
            if self.params is not None:
                self.params = {key: self.params[key].to(dtype) for key in self.params}
            self.dtype = dtype
        return self


    @torch.jit.export
    def clone(
        self,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        """Clone the annaRBM instance.

        Args:
            device (Optional[torch.device], optional): Device. Defaults to None.
            dtype (Optional[torch.dtype], optional): Dtype. Defaults to None.

        Returns:
            annaRBM: annaRBM instance cloned.
        """
        if device is None:
            device = self.device
        if dtype is None:
            dtype = self.dtype
            
        # Cannot instantiate abstract class annaRBM, so use type(self) to instantiate the concrete subclass
        return type(self)(
            params=self.params,
            device=device,
            dtype=dtype,
        )
        

    def save(
        self,
        filename: str,
        num_updates: int,
    ) -> None:
        """Save the parameters of the annaRBM.

        Args:
            filename (str): Path to the h5 archive where to store the model.
            num_updates (int): Number of updates performed so far.
        """
        if self.params is None:
            raise ValueError("Model parameters are not initialized.")
        _save_model(
            params=self.params,
            filename=filename,
            num_updates=num_updates,
        )
    
    
    def load(
        self,
        filename: str,
        device: torch.device,
        dtype: torch.dtype,
        index: int | None = None,
        set_rng_state: bool = False,
    ) -> int:
        """Loads the parameters of the annaRBM.

        Args:
            filename (str): Path to the h5 archive.
            device (torch.device): PyTorch device on which to load the parameters.
            dtype (torch.dtype): Dtype for the parameters.
            index (int | None): Index of the machine to load. If None, the last machine is loaded. Defaults to None.
            set_rng_state (bool): Restore the random state at the given epoch (useful to restore training). Defaults to False.

        Returns:
            int: Number of model updates.
        """
        num_updates, self.params = _load_model(
            filename=filename,
            device=device,
            dtype=dtype,
            index=index,
            set_rng_state=set_rng_state,
        )
        self.device = device
        self.dtype = dtype
        
        return num_updates
    
    
    @abstractmethod
    def compute_energy(
        self,
        visible: torch.Tensor,
        hidden: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """Computes the energy of the visible and hidden units of the input data.

        Args:
            visible (torch.Tensor): Visible units.
            hidden (torch.Tensor): Hidden units.

        Returns:
            torch.Tensor: Energy of the data.
        """
        pass
    
    
    @abstractmethod
    def compute_energy_visibles(
        self,
        visible: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """Computes the energy of the visible units.

        Args:
            visible (torch.Tensor): Visible units.

        Returns:
            torch.Tensor: Energy of the visible units.
        """
        pass
    
    
    @abstractmethod
    def compute_energy_hiddens(
        self,
        hidden: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """Computes the energy of the hidden units.

        Args:
            hidden (torch.Tensor): Hidden units.

        Returns:
            torch.Tensor: Energy of the hidden units.
        """
        pass
    
    
    @abstractmethod
    def sample_hiddens(
        self,
        visible: torch.Tensor,
        label: torch.Tensor,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Samples the hidden units given the visible units and labels.

        Args:
            visible (torch.Tensor): Visible units.
            label (torch.Tensor): Labels.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled hidden units and hidden magnetization.
        """
        pass
    
    
    @abstractmethod
    def sample_visibles(
        self,
        hidden: torch.Tensor,
        label: torch.Tensor,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Samples the visible units given the hidden units and labels.

        Args:
            hidden (torch.Tensor): Hidden units.
            label (torch.Tensor): Labels.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled visible units and visible magnetization.
        """
        pass
        
        
    @abstractmethod
    def sample_labels(
        self,
        hidden: torch.Tensor,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Samples the labels given the hidden units.

        Args:
            hidden (torch.Tensor): Hidden units.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor] Sampled labels and label magnetization.
        """
        pass
        
        
    @abstractmethod
    def sample(
        self,
        gibbs_steps: int,
        visible: torch.Tensor,
        label: torch.Tensor,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Samples from the annaRBM.

        Args:
            gibbs_steps (int): Number of Alternate Gibbs Sampling steps.
            visible (torch.Tensor): Visible units initial configuration.
            label (torch.Tensor): Labels initial configuration.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled visible, hidden and label units.
        """
        pass
    
    
    @abstractmethod
    def sample_conditioned(
        self,
        gibbs_steps: int,
        chains: Dict[str, torch.Tensor] | torch.Tensor | np.ndarray,
        targets: torch.Tensor | np.ndarray,
        beta: float = 1.0,
        **kwargs,
    ) -> torch.Tensor:
        """Samples from the annaRBM conditioned on the target labels. During the sampling, the labels are kept fixed
        and the visible and hidden units are sampled alternatively. The visible's conditional probability distribution
        is returned.

        Args:
            gibbs_steps (int): Number of Alternate Gibbs Sampling steps.
            chains (Dict[str, torch.Tensor] | torch.Tensor | np.ndarray): Chains initialization. It can be either a chain dictionary
                instance or a torch.Tensor (np.ndarray) representing the visible units.
            targets (torch.Tensor | np.ndarray): Target labels.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            torch.Tensor: Conditional probability distribution of the visible units.
        """
        pass
        
        
    def predict_labels(
        self,
        gibbs_steps: int,
        chains: Dict[str, torch.Tensor] | torch.Tensor | np.ndarray,
        targets: torch.Tensor | np.ndarray,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Samples from the annaRBM conditioned on the target visible units. During the sampling, the visibles are kept fixed
        and the labels and hidden units are sampled alternatively. The label's conditional probability distribution is returned.

        Args:
            gibbs_steps (int): Number of Alternate Gibbs Sampling steps.
            chains (Dict[str, torch.Tensor] | torch.Tensor | np.ndarray): Chains initialization. It can be either a chain dictionary
                instance or a torch.Tensor (np.ndarray) representing the labels.
            targets (torch.Tensor | np.ndarray): Target visible units.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Labels's probability distribution.
        """
        pass


    @abstractmethod
    def update_weights_AIS(
        self,
        prev_model: "annaRBM",
        chains: Dict[str, torch.Tensor],
        log_weights: torch.Tensor,
    ) -> torch.Tensor:
        """Update the weights used during the trajectory Annealed Importance Sampling (AIS) algorithm.

        Args:
            prev_model (annRBM): Model at time t-1.
            chains (torch.Tensor): Chains at time t-1.
            log_weights (torch.Tensor): Log-weights at time t-1.

        Returns:
            torch.Tensor: Log-weights at time t.
        """
        pass
    
    
    @abstractmethod
    def compute_log_likelihood(
        self,
        visible: torch.Tensor,
        label: torch.Tensor,
        logZ: float,
        **kwargs,
    ) -> float:
        """Computes the log-likelihood of the model on the given configuration.

        Args:
            visible (torch.Tensor): Visible units.
            label (torch.Tensor): Labels.
            logZ (float): Log partition function.

        Returns:
            float: Log-likelihood of the model on the given configuration.
        """
        pass


    @abstractmethod
    def init_chains(
        self,
        num_samples: int,
    ) -> Dict[str, torch.Tensor]:
        """Initializes the chains.

        Args:
            num_samples (int): Number of samples.

        Returns:
            Dict[str, torch.Tensor]: Initialized chains.
        """
        pass
    

    @staticmethod
    @abstractmethod
    def save_chains(
        filename: str,
        visible: torch.Tensor,
        label: torch.Tensor,
    ) -> None:
        """Save the chains to a file.

        Args:
            filename (str): Path to the file where to store the chains.
            visible (torch.Tensor): Visible units.
            label (torch.Tensor): Labels.
        """
        pass
    
    
    @abstractmethod
    def load_chains(
        self,
        filename: str,
        device: torch.device,
        dtype: torch.dtype,
    ) -> Dict[str, torch.Tensor]:
        """Load the persistent chains from a fasta file.

        Args:
            filename (str): Path to the fasta file.
            device (torch.device): Device on which to load the chains.
            dtype (torch.dtype): Data type for the chains.

        Returns:
            Dict[str, torch.Tensor]: Visible and hidden units and labels of the chains.
        """
        pass        
        
        
    @abstractmethod
    def compute_gradient(
        self,
        data: Dict[str, torch.Tensor],
        chains: Dict[str, torch.Tensor],
        pseudo_count: float = 0.0,
        centered: bool = True,
        lambda_l1: float = 0.0,
        lambda_l2: float = 0.0,
        eta: float = 1.0
    ) -> None:
        """Computes the gradient of the log-likelihood and stores it.

        Args:
            data (Dict[str, torch.Tensor]): Data batch.
            chains (Dict[str, torch.Tensor]): Chains.
            pseudo_count (float, optional): Pseudo count to be added to the data frequencies. Defaults to 0.0.
            centered (bool, optional): Centered gradient. Defaults to True.
            lambda_l1 (float, optional): L1 regularization weight. Defaults to 0.0.
            lambda_l2 (float, optional): L2 regularization weight. Defaults to 0.0.
            eta (float, optional): Relative contribution of the label term. Defaults to 1.0.
        """
        pass
        

    @abstractmethod
    def init_parameters(
        self,
        dataset: annaDataset,
        num_hiddens: int,
        device: torch.device,
        dtype: torch.dtype,
        sigma: float = 1e-4,
    ):
        """Initializes the parameters of the annaRBM.

        Args:
            dataset (aiDataset): Dataset.
            num_hiddens (int): Number of hidden units.
            device (torch.device): Device.
            dtype (torch.dtype): Dtype.
            sigma (float, optional): Standard deviation of the weight matrix. Defaults to 1e-4.

        Returns:
            annaRBM: annaRBM instance.
        """
        pass
        
    
    @abstractmethod
    def num_visibles(self) -> int:
        """Returns the number of visible units.

        Returns:
            int: Number of visible units.
        """
        if self.params is None:
            raise ValueError("Model parameters are not initialized.")
        return self.params["vbias"].shape[0]
        
        
    @abstractmethod
    def num_hiddens(self) -> int:
        """Returns the number of hidden units.

        Returns:
            int: Number of hidden units.
        """
        if self.params is None:
            raise ValueError("Model parameters are not initialized.")
        return self.params["hbias"].shape[0]
        
        
    @abstractmethod
    def num_classes(self) -> int:
        """Returns the number of labels.

        Returns:
            int: Number of labels.
        """
        if self.params is None:
            raise ValueError("Model parameters are not initialized.")
        return self.params["lbias"].shape[0]
        
        
    @abstractmethod
    def logZ0(self) -> float:
        """Computes the initial log partition function for the annaRBM.

        Returns:
            float: Initial log partition function.
        """
        pass
    



import argparse

def add_args_annadca(parser : argparse.ArgumentParser) -> argparse.ArgumentParser:
    dca_args = parser.add_argument_group("Training an annaRBM model.")
    dca_args.add_argument("-d", "--data",         type=str,  required=True,         help="Filename of the dataset to be used for training the model. Can be a fasta file or a csv file containing the sequences and their annotations.")
    dca_args.add_argument("-o", "--output",       type=str,  default="annaRBM",     help="(Defaults to ./annaRBM). Path to the folder where to save the model.")
    dca_args.add_argument("-a", "--annotations",  type=str,  default=None,          help="Path to the file containing the annotations of the sequences.")
    dca_args.add_argument("-H", "--hidden",       type=int,   default=100,          help="(Defaults to 100). Number of hidden units.")
    dca_args.add_argument("--column_names",       type=str,   default="name",       help="(Defaults to name). Name of the column containing the names of the sequences in the annotations file.")
    dca_args.add_argument("--column_sequences",   type=str,   default="sequence",   help="(Defaults to sequence). Name of the column containing the sequences in the annotations file.")
    dca_args.add_argument("--column_labels",      type=str,   default="label",      help="(Defaults to label). Name of the column containing the labels in the annotations file.")
    dca_args.add_argument("-p", "--path_params",  type=str,  default=None,          help="(Defaults to None) Path to the file containing the model's parameters. Required for restoring the training.")
    dca_args.add_argument("-c", "--path_chains",  type=str,  default=None,          help="(Defaults to None) Path to the fasta file containing the model's chains. Required for restoring the training.")
    dca_args.add_argument("-l", "--label",        type=str,   default=None,         help="(Defaults to None). If provoded, adds a label to the output files inside the output folder.")
    dca_args.add_argument("--alphabet",           type=str,   default="protein",    help="(Defaults to protein). Type of encoding for the sequences. Choose among ['protein', 'rna', 'dna'] or a user-defined string of tokens.")
    dca_args.add_argument("--lr",                 type=float, default=0.01,         help="(Defaults to 0.01). Learning rate.")
    dca_args.add_argument("--gibbs_steps",        type=int,   default=10,           help="(Defaults to 10). Number of Alternating Gibbs steps for each gradient estimation.")
    dca_args.add_argument("--nchains",            type=int,   default=5000,         help="(Defaults to 5000). Number of Markov chains to run in parallel. It also corresponds to the batch size.")
    dca_args.add_argument("--nepochs",            type=int,   default=10000,        help="(Defaults to 10000). Maximum number of gradient updates allowed.")
    dca_args.add_argument("--pseudocount",        type=float, default=1e-8,         help="(Defaults to 1e-8). Pseudo count for the single and two-sites statistics. Acts as a regularization.")
    dca_args.add_argument("--eta",                type=float, default=1.0,          help="(Defaults to 1.0). Relative contribution of the label term.")
    dca_args.add_argument("--l1",                 type=float, default=0.0,          help="(Defaults to 0.0). L1 regularization strength.")
    dca_args.add_argument("--l2",                 type=float, default=0.0,          help="(Defaults to 0.0). L2 regularization strength.")
    dca_args.add_argument("--uncentered",         action="store_true",              help="(Defaults to False). If specified, the uncentered version of the gradient is used.")
    dca_args.add_argument("--init_from_profile",  action="store_true",              help="(Defaults to False). If specified, the biases are initialized using the independent-site model.")
    dca_args.add_argument("--is_binary",          action="store_true",              help="(Defaults to False). If specified, the sequences are assumed to be binary (0/1) and the model is trained accordingly.")
    dca_args.add_argument("--seed",               type=int,   default=0,            help="(Defaults to 0). Seed for the random number generator.")
    dca_args.add_argument("--device",             type=str,   default="cuda",       help="(Defaults to cuda). Device to be used.")
    dca_args.add_argument("--dtype",              type=str,   default="float32",    help="(Defaults to float32). Data type to be used.", choices=["float16", "float32", "float64"])
    
    return parser

def add_args_reweighting(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    reweighting_args = parser.add_argument_group("Sequence reweighting arguments")
    reweighting_args.add_argument("-w", "--weights",      type=str,  default=None,  help="(Defaults to None). Path to the file containing the weights of the sequences. If None, the weights are computed automatically.")
    reweighting_args.add_argument("--no_reweighting",     action="store_true",       help="(Defaults to False). If provided, the reweighting of the sequences is not performed.")
    reweighting_args.add_argument("--clustering_seqid",   type=float, default=0.8,   help="(Defaults to 0.8). Sequence Identity threshold for clustering. Used only if 'weights' is not provided.")
    
    return parser

def add_args_train(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    parser = add_args_annadca(parser)
    parser = add_args_reweighting(parser)
    
    return parser


from typing import Dict, Tuple

import h5py
import numpy as np
import torch
from adabmDCA.fasta import write_fasta, get_tokens
import os

from annadca.utils import get_saved_updates

def _save_model(
    params: Dict[str, torch.Tensor],
    filename: str,
    num_updates: int,
):
    """Save the current state of the model.

    Args:
        params (Dict[str, torch.Tensor]): Parameters of the model.
        filename (str): Path to the h5 archive where to store the model.
        num_updates (int): Number of updates performed so far.
    """
    # create the file if it does not exist
    if not os.path.exists(filename):
        with h5py.File(filename, "w") as f:
            pass
    else:
        with h5py.File(filename, "a") as f:
            checkpoint = f.create_group(f"update_{num_updates}")
    
            # Save the parameters of the model
            checkpoint["vbias"] = params["vbias"].detach().cpu().numpy()
            checkpoint["hbias"] = params["hbias"].detach().cpu().numpy()
            checkpoint["lbias"] = params["lbias"].detach().cpu().numpy()
            checkpoint["weight_matrix"] = params["weight_matrix"].detach().cpu().numpy()
            checkpoint["label_matrix"] = params["label_matrix"].detach().cpu().numpy()
            
    
            # Save current random state
            checkpoint["torch_rng_state"] = torch.get_rng_state()
            checkpoint["numpy_rng_arg0"] = np.random.get_state()[0]
            checkpoint["numpy_rng_arg1"] = np.random.get_state()[1]
            checkpoint["numpy_rng_arg2"] = np.random.get_state()[2]
            checkpoint["numpy_rng_arg3"] = np.random.get_state()[3]
            checkpoint["numpy_rng_arg4"] = np.random.get_state()[4]
    

def _load_model(
    filename: str,
    device: torch.device,
    dtype: torch.dtype,
    index: int | None = None,
    set_rng_state: bool = False,
) -> Tuple[int, Dict[str, torch.Tensor]]:
    """Loads a RBM from an h5 archive.

    Args:
        filename (str): Path to the h5 archive.
        device (torch.device): PyTorch device on which to load the parameters and the chains.
        dtype (torch.dtype): Dtype for the parameters and the chains.
        index (int | None, optional): Index of the machine to load. If None, the last machine is loaded. Defaults to None.
        set_rng_state (bool, optional): Restore the random state at the given epoch (useful to restore training). Defaults to False.

    Returns:
        Tuple[int, Dict[str, torch.Tensor]]: Number of updates and parameters of the loaded model.
    """
    list_updates = get_saved_updates(filename)
    if index is None:
        index = list_updates[-1]
    else:
        if index not in list_updates:
            raise ValueError(f"Index {index} not found in the h5 archive.")
    
    last_file_key = f"update_{index}"
    with h5py.File(filename, "r") as f:
        weight_matrix = torch.tensor(
            f[last_file_key]["weight_matrix"][()],
            device=device,
            dtype=dtype,
        )
        label_matrix = torch.tensor(
            f[last_file_key]["label_matrix"][()],
            device=device,
            dtype=dtype,
        )
        vbias = torch.tensor(f[last_file_key]["vbias"][()], device=device, dtype=dtype)
        hbias = torch.tensor(f[last_file_key]["hbias"][()], device=device, dtype=dtype)
        lbias = torch.tensor(f[last_file_key]["lbias"][()], device=device, dtype=dtype)
        
        if set_rng_state:
            torch.set_rng_state(torch.tensor(np.array(f[last_file_key]["torch_rng_state"])))
            np_rng_state = tuple(
                [
                    f[last_file_key]["numpy_rng_arg0"][()].decode("utf-8"),
                    f[last_file_key]["numpy_rng_arg1"][()],
                    f[last_file_key]["numpy_rng_arg2"][()],
                    f[last_file_key]["numpy_rng_arg3"][()],
                    f[last_file_key]["numpy_rng_arg4"][()],
                ]
            )
            np.random.set_state(np_rng_state)
        
    params = {
        "weight_matrix": weight_matrix,
        "label_matrix": label_matrix,
        "vbias": vbias,
        "hbias": hbias,
        "lbias": lbias,
    }
    
    return index, params


def _save_chains(
    filename: str,
    visible: torch.Tensor,
    label: torch.Tensor,
    alphabet: str,
) -> None:
    """Save the persistent chains on a fasta file.
    
    Args:
        filename (str): Path to the fasta file.
        visible (torch.Tensor): Visible units of the chains.
        label (torch.Tensor): Labels of the chains.
        alphabet (str): Alphabet to be used for the encosing of the sequences.
    """
    tokens = get_tokens(alphabet)
    visible = visible.int().cpu().numpy()
    label = label.int().cpu().numpy()
    # Headers are associated with the labels
    headers = np.vectorize(lambda x: "".join([str(i) for i in x]), signature="(l) -> ()")(label)
    write_fasta(
        fname=filename,
        headers=headers,
        sequences=visible,
        tokens=tokens,
        remove_gaps=False,
    )




from typing import Dict, Tuple
import torch
from adabmDCA.functional import one_hot


@torch.jit.script
def _sample_hiddens(
    visible: torch.Tensor,
    label: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float = 1.0,
) -> Tuple[torch.Tensor, torch.Tensor]:

    L, q, P = params["weight_matrix"].shape
    weight_matrix_oh = params["weight_matrix"].view(L * q, P)
    visible_oh = visible.view(-1, L * q)
    hidden_mag = torch.sigmoid(
        beta * (params["hbias"] + (visible_oh @ weight_matrix_oh) + (label @ params["label_matrix"]))
    )
    hidden = torch.bernoulli(hidden_mag).to(params["weight_matrix"].dtype)
    return (hidden, hidden_mag)


@torch.jit.script
def _sample_visibles(
    hidden: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float = 1.0,
) -> Tuple[torch.Tensor, torch.Tensor]:

    num_visibles, num_states, _ = params["weight_matrix"].shape
    visible_mag = torch.softmax(
        beta * (params["vbias"] + torch.einsum("np,lqp->nlq", hidden, params["weight_matrix"])),
        dim=-1
    )
    visible = one_hot(
        torch.multinomial(visible_mag.view(-1, num_states), 1).view(-1, num_visibles),
        num_classes=num_states,
    ).to(params["weight_matrix"].dtype)
    return (visible, visible_mag)


@torch.jit.script
def _sample_labels(
    hidden: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float = 1.0,
) -> Tuple[torch.Tensor, torch.Tensor]:

    label_mag = torch.softmax(
        beta * (params["lbias"] + (hidden @ params["label_matrix"].T)),
        dim=-1,
    )
    label = one_hot(torch.multinomial(label_mag, 1), num_classes=label_mag.shape[-1]).to(params["weight_matrix"].dtype).view(-1, label_mag.shape[-1])
    
    return (label, label_mag)


def _sample(
    gibbs_steps: int,
    visible: torch.Tensor,
    label: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float = 1.0,
    **kwargs,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:

    hidden = torch.zeros((visible.shape[0], params["hbias"].shape[0]), device=visible.device, dtype=visible.dtype)
    hidden_mag = torch.zeros_like(hidden)
    visible_mag = torch.zeros_like(visible)
    for _ in range(gibbs_steps):
        hidden, hidden_mag = _sample_hiddens(visible=visible, label=label, params=params, beta=beta)
        label, label_mag = _sample_labels(hidden=hidden, params=params, beta=beta)
        visible, visible_mag = _sample_visibles(hidden=hidden, params=params, beta=beta)

    return (visible, hidden, label, visible_mag, hidden_mag, label_mag)


def _sample_conditioned(
    gibbs_steps: int,
    visible: torch.Tensor,
    label: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float = 1.0,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    hidden = torch.zeros((visible.shape[0], params["hbias"].shape[0]), device=visible.device, dtype=visible.dtype)
    hidden_mag = torch.zeros_like(hidden)
    visible_mag = torch.zeros_like(visible)
    for _ in range(gibbs_steps):
        hidden, hidden_mag = _sample_hiddens(visible=visible, label=label, params=params, beta=beta)
        visible, visible_mag = _sample_visibles(hidden=hidden, params=params, beta=beta)

    return (visible, hidden, visible_mag, hidden_mag)


@torch.jit.script
def _predict_labels(
    visible: torch.Tensor,
    params: Dict[str, torch.Tensor],
    beta: float = 1.0,    
) -> torch.Tensor:
    """Returns the probability of each label given the visible units: p(l|v).

    Args:
        visible (torch.Tensor): Visible units.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        beta (float, optional): Inverse temperature. Defaults to 1.0.

    Returns:
        torch.Tensor: p(l|v).
    """
    if visible.dim() == 2:
        visible = visible.unsqueeze(0)
    num_visibles, num_states, num_hiddens = params["weight_matrix"].shape
    num_labels = params["label_matrix"].shape[0]
    
    v_flat = visible.view(visible.shape[0], -1)
    w_oh = params["weight_matrix"].view(num_visibles * num_states, num_hiddens)
    exponent = (
        params["hbias"].view(1, 1, num_hiddens) +
        (v_flat @ w_oh).view(-1, 1, num_hiddens) +
        params["label_matrix"].view(1, num_labels, num_hiddens)
    )
    logits = params["lbias"] + torch.where(exponent > 10, exponent, torch.log1p(torch.exp(exponent))).sum(-1) # (num_samples, num_labels)
    
    return torch.softmax(beta * logits, dim=-1)


import argparse
import os
import numpy as np
from tqdm import tqdm
from itertools import cycle
import time
import warnings

import torch
from torch.utils.data import DataLoader
from torch.optim import SGD
from adabmDCA.utils import get_device, get_dtype
from adabmDCA.stats import get_freq_single_point as get_freq_single_point_cat

from annadca.parser import add_args_train
from annadca.dataset import annaDataset
from annadca import annaRBMbin, annaRBMcat
from annadca.rbm.binary.stats import get_freq_single_point as get_freq_single_point_bin
from annadca.train import pcd
from annadca.utils import get_saved_updates


# import command-line input arguments
def create_parser():
    # Important arguments
    parser = argparse.ArgumentParser(description='Train an annaRBM model.')
    parser = add_args_train(parser) 
    
    return parser


if __name__ == '__main__':
    
    # Load parser, training dataset and DCA model
    parser = create_parser()
    args = parser.parse_args()
    
    print("\n" + "".join(["*"] * 10) + f" Training annaRBM model " + "".join(["*"] * 10) + "\n")
    device = get_device(args.device)
    dtype = get_dtype(args.dtype)
    template = "{0:<30} {1:<50}"
    print(template.format("Input MSA:", str(args.data)))
    print(template.format("Input annotations:", str(args.annotations)))
    print(template.format("Output folder:", str(args.output)))
    print(template.format("Number of hidden units:", args.hidden))
    print(template.format("Learning rate:", args.lr))
    print(template.format("Minibatch size:", args.nchains))
    print(template.format("Number of chains:", args.nchains))
    print(template.format("Number of Gibbs steps:", args.gibbs_steps))
    print(template.format("Number of epochs:", args.nepochs))
    print(template.format("Centered gradient:", not args.uncentered))
    print(template.format("L1 regularization:", args.l1))
    print(template.format("L2 regularization:", args.l2))
    print(template.format("Profile initialization:", args.init_from_profile))
    print(template.format("Labels contribution:", args.eta))
    if args.pseudocount is not None:
        print(template.format("Pseudocount:", args.pseudocount))
    print(template.format("Random seed:", args.seed))
    print(template.format("Data type:", args.dtype))
    print("\n")

    # Check that input files exist
    for path in [args.data, args.annotations, args.path_params, args.path_chains]:
        if path is not None and not os.path.exists(path):
            raise FileNotFoundError(f"Input file {path} not found.")

    # Import data
    print("Importing dataset...")
    dataset = annaDataset(
        path_data=args.data,
        path_ann=args.annotations,
        column_names=args.column_names,
        column_sequences=args.column_sequences,
        column_labels=args.column_labels,
        is_binary=args.is_binary,
        alphabet=args.alphabet,
        clustering_th=args.clustering_seqid,
        no_reweighting=args.no_reweighting,
        path_weights=args.weights,
        device=device,
        dtype=dtype,
    )
    tokens = dataset.tokens
    print(f"Alphabet: {dataset.tokens}")
    print(f"Dataset imported successfully: M={len(dataset)}, L={dataset.L}, q={dataset.q}.")
    
    # Create the folder where to save the model
    folder = args.output
    os.makedirs(folder, exist_ok=True)

    if args.label is not None:
        file_paths = {
            "log" : os.path.join(folder, f"{args.label}.log"),
            "params" : os.path.join(folder, f"{args.label}_params.h5"),
            "chains" : os.path.join(folder, f"{args.label}_chains.fasta")
        }
        
    else:
        file_paths = {
            "log" : os.path.join(folder, f"annaRBM.log"),
            "params" : os.path.join(folder, f"params.h5"),
            "chains" : os.path.join(folder, f"chains.fasta")
        }
        
    # Check if the files in file_paths already exist. If so, delete them
    if args.path_params is None:
        for path in file_paths.values():
            if os.path.exists(path):
                # Ask for confirmation before deleting
                confirm = input(f"File {path} already exists. Do you want to delete it? (y/n): ")
                if confirm.lower() == "y":
                    os.remove(path)

    # Save the weights if not already provided
    if args.weights is None and not args.no_reweighting:
        if args.label is not None:
            path_weights = os.path.join(folder, f"{args.label}_weights.dat")
        else:
            path_weights = os.path.join(folder, "weights.dat")
        np.savetxt(path_weights, dataset.weights.cpu().numpy())
        print(f"Weights saved in {path_weights}")
    elif args.no_reweighting:
        print("All sequence weights set to 1.0")
        
    # Set the random seed
    torch.manual_seed(args.seed)
        
    # Shuffle the dataset
    dataset.shuffle()
    
    if args.pseudocount is None:
        args.pseudocount = 1. / dataset.get_effective_size()
        print(f"Pseudocount automatically set to {args.pseudocount}.")
        
    # Initialize the model and the chains
    num_visibles = dataset.get_num_residues()
    num_hiddens = args.hidden
    num_labels = dataset.get_num_classes()
    num_states = dataset.get_num_states()
    if dataset.is_binary:
        rbm = annaRBMbin()
        data = dataset.data_one_hot
        frequences_visible = get_freq_single_point_bin(
            data=data,
            weights=dataset.weights,
            pseudo_count=args.pseudocount,
        )
    elif not dataset.is_binary:
        rbm = annaRBMcat()
        data = dataset.data_one_hot
        frequences_visible = get_freq_single_point_cat(
            data=data,
            weights=dataset.weights,
            pseudo_count=args.pseudocount,
        )
    frequency_labels = get_freq_single_point_bin(
        data=dataset.labels_one_hot,
        weights=dataset.weights,
        pseudo_count=args.pseudocount,
    )

    if args.path_params is not None:
        upd = rbm.load(
            filename=args.path_params,
            device=device,
            dtype=dtype,
            set_rng_state=True,
        )
        print("Model parameters loaded from", args.path_params)
    else:
        if args.init_from_profile:
            init_frequences_visible = frequences_visible
            init_frequences_labels = frequency_labels
        else:
            init_frequences_visible = None
            init_frequences_labels = None
            
        rbm.init_parameters(
            num_visibles=num_visibles,
            num_hiddens=num_hiddens,
            num_labels=num_labels,
            num_states=num_states,
            frequencies_visibles=init_frequences_visible,
            frequencies_labels=init_frequences_labels,
            std_init=1e-4,
            device=device,
            dtype=dtype,
        )
        upd = 0
    
    if args.path_chains is not None:
        chains = rbm.load_chains(
            filename=args.path_chains,
            device=device,
            dtype=dtype,
            alphabet=tokens,
        )
        print("Chains loaded from", args.path_chains)
    else:
        if args.nchains >= dataset.__len__():
            args.nchains = dataset.__len__()
            warnings.warn("The number of chains is larger than the dataset size. The number of chains is set to the dataset size.")
        chains = rbm.init_chains(num_samples=args.nchains, use_profile=True)
        
    print("\n")
    # Save the hyperparameters of the model
    template = "{0:<20} {1:<10}\n"  
    with open(file_paths["log"], "w") as f:
        if args.label is not None:
            f.write(template.format("label:", args.label))
        else:
            f.write(template.format("label:", "N/A"))
            
        f.write(template.format("input data:", str(args.data)))
        f.write(template.format("alphabet:", dataset.tokens))
        f.write(template.format("# hiddens:", args.hidden))
        f.write(template.format("nchains:", args.nchains))
        f.write(template.format("minibatch size:", args.nchains))
        f.write(template.format("gibbs steps:", args.gibbs_steps))
        f.write(template.format("lr:", args.lr))
        f.write(template.format("pseudo count:", args.pseudocount))
        f.write(template.format("centered:", not args.uncentered))
        f.write(template.format("profile init:", args.init_from_profile))
        f.write(template.format("l1 strength:", args.l1))
        f.write(template.format("l2 strength:", args.l2))
        f.write(template.format("eta:", args.eta))
        f.write(template.format("random seed:", args.seed))
        f.write("\n")
        template = "{0:<10} {1:<10}\n"
        f.write(template.format("Epoch", "Time [s]"))
        
    # Initialize gradients for the parameters
    for key, value in rbm.params.items():
        value.grad = torch.zeros_like(value)

    # Select the optimizer
    optimizer = SGD(rbm.params.values(), lr=args.lr, maximize=True)
    
    # Initialize the dataloader
    dataloader = DataLoader(
        dataset,
        batch_size=args.nchains,
        shuffle=True,
        drop_last=True,
    )
    
    # Allows to iterate indefinitely on the dataloader without worrying on the epochs
    dataloader = cycle(dataloader)
    
    # Initialize the variables for the log-likelihood estimation
    logZ = rbm.logZ0()
    log_weights = torch.zeros(args.nchains, device=device, dtype=dtype)
    log_likelihood = rbm.compute_log_likelihood(
        visible=data,
        label=dataset.labels_one_hot,
        weight=dataset.weights,
        logZ=logZ,
    )
    
    # Train the model
    start = time.time()
    pbar = tqdm(initial=upd, total=args.nepochs, colour="red", dynamic_ncols=True, ascii="-#")
    if upd == 0:
        rbm.save(filename=file_paths["params"], num_updates=upd)
    with torch.no_grad():
        while upd < args.nepochs:
            upd += 1
            if upd % 10 == 0:
                pbar.update(10)

            # Get the next batch
            batch = next(dataloader)
            optimizer.zero_grad(set_to_none=False)
            chains = pcd(
                rbm=rbm,
                data_batch=batch,
                chains=chains,
                gibbs_steps=args.gibbs_steps,
                pseudo_count=args.pseudocount,
                centered=(not args.uncentered),
                lambda_l1=args.l1,
                lambda_l2=args.l2,
                eta=args.eta,
            )

            # Update the parameters
            optimizer.step()
            
            # Set the gauge of the weights
            rbm.zerosum_gauge()

            if upd % 1000 == 0:
                rbm.save(
                    filename=file_paths["params"],
                    num_updates=upd,
                )
                rbm.save_chains(
                    filename=file_paths["chains"],
                    visible=chains["visible"],
                    label=chains["label"],
                    alphabet=dataset.tokens,
                )
                with open(file_paths["log"], "a") as f:
                    f.write(template.format(f"{upd}", f"{time.time() - start:.2f}"))
        pbar.close()
        
        # Save the final model and chains if upd is not present in the h5 archive
        saved_updates = get_saved_updates(file_paths["params"])
        if upd not in saved_updates:
            rbm.save(
                filename=file_paths["params"],
                num_updates=upd,
            )
            rbm.save_chains(
                filename=file_paths["chains"],
                visible=chains["visible"],
                label=chains["label"],
                alphabet=dataset.tokens,
            )
            with open(file_paths["log"], "a") as f:
                f.write(template.format(f"{upd}", f"{time.time() - start:.2f}"))


from typing import Dict
import torch


def _compute_energy(
    visible: torch.Tensor,
    hidden: torch.Tensor,
    label: torch.Tensor,
    params: Dict[str, torch.Tensor],
    **kwargs,
) -> torch.Tensor:
    def _compute_energy_chain(
        visible: torch.Tensor,
        hidden: torch.Tensor,
        label: torch.Tensor,
        params: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        fields = (visible @ params["vbias"]) + (hidden @ params["hbias"]) + (label @ params["lbias"])
        interaction = (visible @ params["weight_matrix"] @ hidden) + (label @ params["label_matrix"] @ hidden)
        return - fields - interaction
    
    return torch.vmap(_compute_energy_chain, in_dims=(0, 0, 0, None))(visible, hidden, label, params)


def _compute_energy_visibles_labels(
    visible: torch.Tensor,
    label: torch.Tensor,
    params: Dict[str, torch.Tensor],
    **kwargs,
) -> torch.Tensor:
    @torch.jit.script
    def _compute_energy_chain(
        visible: torch.Tensor,
        label: torch.Tensor,
        params: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        fields = (visible @ params["vbias"]) + (label @ params["lbias"])
        exponent = params["hbias"] + (visible @ params["weight_matrix"]) + (label @ params["label_matrix"])
        log_term = torch.where(exponent < 10, torch.log(1.0 + torch.exp(exponent)), exponent)
        return - fields - log_term.sum()
    
    return torch.vmap(_compute_energy_chain, in_dims=(0, 0, None))(visible, label, params)


def _compute_energy_visibles(
    visible: torch.Tensor,
    params: Dict[str, torch.Tensor],
    **kwargs,
) -> torch.Tensor:

    def _compute_energy_chain(
        visible: torch.Tensor,
        params: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        fields = visible @ params["vbias"]
        exponent = params["hbias"].unsqueeze(0) + \
            + (visible @ params["weight_matrix"]).unsqueeze(0) + params["label_matrix"] # (num_classes, num_hiddens)
        log_term = torch.where(exponent < 10, torch.log(1.0 + torch.exp(exponent)), exponent)
        exponent_label_term = log_term.sum(1) + params["lbias"] # (num_classes,)
        return - fields - torch.logsumexp(exponent_label_term, dim=0)

    return torch.vmap(_compute_energy_chain, in_dims=(0, None))(visible, params)


def _compute_energy_hiddens(
    hidden: torch.Tensor,
    params: Dict[str, torch.Tensor],
    **kwargs,
) -> torch.Tensor:

    @torch.jit.script
    def _compute_energy_chain(
        hidden: torch.Tensor,
        params: Dict[str, torch.Tensor],
    ) -> torch.Tensor:
        # Effective parameters as if the labels were visible units
        eff_bias = torch.hstack([params["vbias"], params["lbias"]])
        eff_weight_matrix = torch.vstack([params["weight_matrix"], params["label_matrix"]])
        field = hidden @ params["hbias"]
        exponent = eff_bias + (hidden @ eff_weight_matrix.T)
        log_term = torch.where(exponent < 10, torch.log(1.0 + torch.exp(exponent)), exponent)
        return - field - log_term.sum()
    
    return torch.vmap(_compute_energy_chain, in_dims=(0, None))(hidden, params)


def _update_weights_AIS(
    prev_params: Dict[str, torch.Tensor],
    curr_params: Dict[str, torch.Tensor],
    chains: Dict[str, torch.Tensor],
    log_weights: torch.Tensor,
) -> torch.Tensor:

    energy_prev = _compute_energy(**chains, params=prev_params)
    energy_curr = _compute_energy(**chains, params=curr_params)
    log_weights += energy_prev - energy_curr
    
    return log_weights


def _compute_log_likelihood(
    visible: torch.Tensor,
    weight: torch.Tensor,
    params: Dict[str, torch.Tensor],
    logZ: float,
) -> float:
    
    energy_data = _compute_energy_visibles(
        visible=visible,
        params=params,
    )
    mean_energy_data = (energy_data * weight.view(-1)).sum() / weight.sum()

    return - mean_energy_data.item() - logZ


import numpy as np
from typing import List, Tuple
import h5py
import torch


def _complete_labels(
    dict_labels: dict,
    names: list | np.ndarray,
) -> dict:
    """Goes through 'names' and returns a dictionary of 'name -> label' with 'None' for missing labels or labels
    marked as 'None' in 'dict_labels'.
    
    Args:
        dict_labels (dict): Dictionary of 'name -> label'.
        names (list | np.ndarray): List of names to complete.
        
    Returns:
        dict: Dictionary of 'name -> label' with 'None' for missing labels or labels marked as 'None' in 'dict_labels'.
    """
    name_to_label = {}
    for n in names:
        if n not in dict_labels:
            name_to_label[n] = None
        elif dict_labels[n] == "None":
            name_to_label[n] = None
        else:
            name_to_label[n] = dict_labels[n]
    
    return name_to_label


def _encode_labels(
    dict_labels: dict,
    start_idx: int,
) -> Tuple[dict, np.array]:
    """Take a dictionary of 'name -> label' and return a dictionary of 'label -> one-hot index' and a one-hot encoded array of labels.
    When the label is marked as 'None', it is represented with an array of zeros.

    Args:
        dict_labels (dict): name to label dictionary.
        start_idx (int): Starting index for the one-hot encoding. Used if multiple categorizations are needed.

    Returns:
        Tuple[dict, np.array]: Dictionary of label to index in the one-hot encoding and one-hot encoded array of labels.
    """
    # Get all unique labels that are not None
    unique_labels = np.unique([lab for lab in dict_labels.values() if lab != None])
    num_categ = len(unique_labels)
    # None label is represented as an array of zeros
    one_hot_transform = np.vstack([np.zeros(num_categ), np.eye(num_categ)])
    label_to_index = {lab.item() : (i + 1) for i, lab in enumerate(unique_labels)}
    label_to_index[None] = 0
    numeric_labels = np.array([label_to_index[n] for n in dict_labels.values()])
    one_hot_labels = one_hot_transform[numeric_labels]
    # Shift index to start_idx and subtract 1 to account for the None label
    label_to_index = {k: (v - 1 + start_idx) for k, v in label_to_index.items() if k != None}
    
    return label_to_index, one_hot_labels


def _parse_labels(labels_dict_list: List[dict]) -> Tuple[dict, np.array]:
    """Parse a list of dictionaries of labels and return a dictionary of label to index and a one-hot encoded array of labels.

    Args:
        labels_dict_list (List[dict]): List of dictionaries of labels.

    Returns:
        Tuple[dict, np.array]: Dictionary of label to index in the one-hot encoding and one-hot encoded array of labels.
    """
    start_idx = 0
    label_to_idx = {}
    one_hot_labels = []
    
    for labels_dict in labels_dict_list:
        label_to_idx_, one_hot_labels_ = _encode_labels(labels_dict, start_idx)
        label_to_idx.update(label_to_idx_)
        one_hot_labels.append(one_hot_labels_)
        start_idx += len(label_to_idx_)
    
    one_hot_labels = np.concatenate(one_hot_labels, axis=1)
    
    return label_to_idx, one_hot_labels


def get_saved_updates(filename: str) -> np.ndarray:
    """Returns the list of indices of the saved updates in the h5 archive.

    Args:
        filename (str): Path to the h5 archive.

    Returns:
        np.ndarray: Array of the indices of the saved updates.
    """
    updates = []
    with h5py.File(filename, "r") as f:
        for key in f.keys():
            if "update" in key:
                update = int(key.replace("update_", ""))
                updates.append(update)
    return np.sort(np.array(updates))


def get_eigenvalues_history(
    filename: str,
    target_matrix: str,
    device: torch.device,
    dtype: torch.dtype = torch.float32,
) -> Tuple[np.ndarray, np.ndarray]:
    """For each update in the file, return the eigenvalues of the target matrix.
    
    Args:
        filename (str): The filename of the h5 file containing the model.
        target_matrix (str): The matrix for which to compute the eigenvalues. Must be either 'weight_matrix' or 'label_matrix'.
        device (torch.device): The device on which to compute the eigenvalues.
        dtype (torch.dtype): The dtype of the eigenvalues.
    
    Returns:
        Tuple[np.ndarray, np.ndarray]: The updates indices and the eigenvalues.
    
    """
    if target_matrix not in ["weight_matrix", "label_matrix"]:
        raise ValueError("target_matrix must be either 'weight_matrix' or 'label_matrix'")
    f = h5py.File(filename, 'r')
    updates = []
    eigenvalues = []
    for key in f.keys():
        if "update" in key:
            matrix = f[key][f"{target_matrix}"][()]
            matrix = torch.tensor(matrix, device=device, dtype=dtype)
            matrix = matrix.reshape(-1, matrix.shape[-1])
            eig = torch.linalg.svdvals(matrix).cpu().numpy()
            eigenvalues.append(eig.reshape(*eig.shape, 1))
            updates.append(int(key.split("_")[1]))
    
    # Sort the results
    sorting = np.argsort(updates)
    updates = np.array(updates)[sorting]
    eigenvalues = np.array(np.hstack(eigenvalues).T)[sorting]
    f.close()
            
    return updates, eigenvalues


def mutual_information(
    visible: torch.Tensor,
    label: torch.Tensor,
    **kwargs,
) -> torch.Tensor:
    """Estimates the mutual information between the visible and the label units.
    
    Args:
        visible (torch.Tensor): Visible units.
        label (torch.Tensor): Label units.
        
    Returns:
        torch.Tensor: The mutual information between the visible and the label units
    """
    nchains = len(visible)
    fl = label.mean(0)
    fi = visible.mean(0)
    fil = label.T @ visible / nchains
    Iil = fil * (torch.log(1e-8 + fil) - torch.log(1e-8 + (fl.unsqueeze(1) @ fi.unsqueeze(0))))
    
    return Iil


from typing import Dict
import torch

from annadca.classes import annaRBM


def pcd(
    rbm: annaRBM,
    data_batch: Dict[str, torch.Tensor],
    chains: Dict[str, torch.Tensor],
    gibbs_steps: int,
    pseudo_count: float = 0.0,
    centered: bool = True,
    lambda_l1: float = 0.0,
    lambda_l2: float = 0.0,
    eta: float = 1.0,
) -> Dict[str, torch.Tensor]:
    """Computes the gradient of the parameters of the model and the Markov chains using the Persistent Contrastive Divergence algorithm.

    Args:
        rbm (annaRBM): RBM model to be trained.
        data_batch (Dict[str, torch.Tensor]): Batch of data.
        chains (Dict[str, torch.Tensor]): Persistent chains.
        gibbs_steps (int): Number of Alternating Gibbs Sampling steps.
        pseudo_count (float, optional): Pseudo count to be added to the data frequencies. Defaults to 0.0.
        centered (bool, optional): Whether to use centered gradients. Defaults to True.
        lambda_l1 (float, optional): L1 regularization weight. Defaults to 0.0.
        lambda_l2 (float, optional): L2 regularization weight. Defaults to 0.0.
        eta (float, optional): Relative contribution of the label term. Defaults to 1.0.

    Returns:
        Dict[str, torch.Tensor]: Updated chains.
    """
    # Compute the hidden magnetization of the data
    data_batch["hidden"] = rbm.sample_hiddens(**data_batch, beta=1.0)["hidden_mag"]

    # Compute the gradient of the Log-Likelihood
    rbm.compute_gradient(
        data=data_batch,
        chains=chains,
        pseudo_count=pseudo_count,
        centered=centered,
        lambda_l1=lambda_l1,
        lambda_l2=lambda_l2,
        eta=eta,
    )
    
    # Update the persistent chains
    chains = rbm.sample(gibbs_steps=gibbs_steps, **chains, beta=1.0)

    return chains


from typing import Optional, Dict
import numpy as np

import torch
from adabmDCA.fasta import import_from_fasta

from annadca.classes import annaRBM
from annadca.io import _save_chains
from annadca.rbm.binary.statmech import (
    _compute_energy,
    _compute_energy_visibles,
    _compute_energy_hiddens,
    _update_weights_AIS,
    _compute_log_likelihood,
)
from annadca.rbm.binary.sampling import (
    _sample,
    _sample_hiddens,
    _sample_visibles,
    _sample_labels,
    _sample_conditioned,
    _predict_labels,
)
from annadca.rbm.binary.init import _init_parameters, _init_chains
from annadca.rbm.binary.grad import _compute_gradient


class annaRBMbin(annaRBM):
    def __init__(
        self,
        params: Dict[str, torch.Tensor] | None = None,
        device: Optional[torch.device] = torch.device("cpu"),
        dtype: Optional[torch.dtype] = torch.float32,
    ):
        super().__init__(params, device, dtype)
        
        
    def to(
        self,
        device: Optional[torch.device] | None = None,
        dtype: Optional[torch.dtype] | None = None,
    ):
        """Moves the parameters to the specified device and/or dtype.

        Args:
            device (Optional[torch.device], optional): Device. Defaults to None.
            dtype (Optional[torch.dtype], optional): Dtype. Defaults to None.

        Returns:
            annaRBM: annaRBM instance with the parameters moved to the specified device and/or dtype.
        """
        return super().to(device, dtype)
    
    
    def clone(
        self,
        device: Optional[torch.device] | None = None,
        dtype: Optional[torch.dtype] | None = None,
    ):
        """Clone the annaRBMbin instance.

        Args:
            device (Optional[torch.device], optional): Device. Defaults to None.
            dtype (Optional[torch.dtype], optional): Dtype. Defaults to None.

        Returns:
            annaRBM: annaRBM instance cloned.
        """
        if device is None:
            device = self.device
        if dtype is None:
            dtype = self.dtype
            
        cloned_params = {k: v.clone().to(device=device, dtype=dtype) for k, v in self.params.items()}
            
        return annaRBMbin(
            params=cloned_params,
            device=device,
            dtype=dtype,
        )
    
    
    def save(
        self,
        filename: str,
        num_updates: int,
    ) -> None:
        """Save the parameters of the annaRBM.

        Args:
            filename (str): Path to the h5 archive where to store the model.
            num_updates (int): Number of updates performed so far.
        """
        return super().save(filename, num_updates)
    
    
    def load(
        self,
        filename: str,
        index: int | None = None,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
        set_rng_state: bool = False,
    ) -> int:
        """Loads the parameters of the annaRBM.

        Args:
            filename (str): Path to the h5 archive.
            index (int): Index of the machine to load.
            device (torch.device): PyTorch device on which to load the parameters.
            dtype (torch.dtype): Dtype for the parameters.
            set_rng_state (bool): Restore the random state at the given epoch (useful to restore training). Defaults to False.

        Returns:
            int: Number of model updates
        """
        return super().load(
            filename=filename,
            index=index,
            device=device,
            dtype=dtype,
            set_rng_state=set_rng_state,
            )
        
        
    def compute_energy(
        self,
        visible: torch.Tensor,
        hidden: torch.Tensor,
        label: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """Computes the energy of the model on the given configuration.

        Args:
            visible (torch.Tensor): Visible units.
            hidden (torch.Tensor): Hidden units.
            label (torch.Tensor): Labels.

        Returns:
            torch.Tensor: Energy of the model on the given configuration.
        """
    
        return _compute_energy(visible, hidden, label, self.params)
    
    
    def compute_energy_visibles(
        self,
        visible: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """Returns the energy of the model computed on the input visibles.

        Args:
            visible (torch.Tensor): Visible units.

        Returns:
            torch.Tensor: Energy of the data points.
        """
        return _compute_energy_visibles(visible, self.params)
    
    
    def compute_energy_hiddens(
        self,
        hidden: torch.Tensor,
        **kwargs,
    ) -> torch.Tensor:
        """Computes the energy of the model on the hidden layer.

        Args:
            hidden (torch.Tensor): Hidden units.

        Returns:
            torch.Tensor: Energy of the model on the hidden layer.
        """
        return _compute_energy_hiddens(hidden, self.params)
    
    
    def sample_hiddens(
        self,
        visible: torch.Tensor,
        label: torch.Tensor,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Sample the hidden layer conditionally to the visible one and the labels.

        Args:
            visible (torch.Tensor): Visible units.
            label (torch.Tensor): Labels.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled hidden units and hidden magnetization.
        """
        hidden, hidden_mag = _sample_hiddens(visible, label, self.params, beta)
        return {"hidden": hidden, "hidden_mag": hidden_mag}
    
    
    def sample_visibles(
        self,
        hidden: torch.Tensor,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Sample the visible layer conditionally to the hidden one.

        Args:
            hidden (torch.Tensor): Hidden units.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled visible units and visible magnetization.
        """
        visible, visible_mag = _sample_visibles(hidden, self.params, beta)
        return {"visible": visible, "visible_mag": visible_mag}
    
    
    def sample_labels(
        self,
        hidden: torch.Tensor,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Sample the labels conditionally to the hidden layer.

        Args:
            hidden (torch.Tensor): Hidden units.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled labels and label magnetization.
        """
        label, label_mag = _sample_labels(hidden, self.params, beta)
        return {"label": label, "label_mag": label_mag}
    
    
    def sample(
        self,
        gibbs_steps: int,
        num_samples: int | None = None,
        visible: torch.Tensor | None = None,
        label: torch.Tensor | None = None,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Samples from the binary annaRBM.

        Args:
            gibbs_steps (int): Number of Gibbs steps.
            num_samples (int | None, optional): Number of samples to generate. If None, it will be inferred from the visible or label tensors. Defaults to None.
            visible (torch.Tensor): Visible units. If None, it will be initialized randomly.
            label (torch.Tensor): Labels. If None, it will be initialized randomly. If both visible and label are None, num_samples must be provided.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled visible, hidden units and labels.
        """
        # Infer num_samples if possible
        if visible is not None:
            num_samples = visible.shape[0]
        elif label is not None:
            num_samples = label.shape[0]
        elif num_samples is None or num_samples <= 0:
            raise ValueError("Either visible, label or a positive num_samples must be provided.")

        # Initialize missing visible/label
        chains_init = self.init_chains(num_samples=num_samples)
        if visible is None:
            visible = chains_init["visible"]
        if label is None:
            label = chains_init["label"]

        if visible.shape[0] != label.shape[0]:
            raise ValueError(f"The number of visible units ({visible.shape[0]}) and labels ({label.shape[0]}) must be the same.")

        visible, hidden, label, visible_mag, hidden_mag, label_mag = _sample(gibbs_steps, visible, label, self.params, beta)
        return {"visible": visible, "hidden": hidden, "label": label, "visible_mag": visible_mag, "hidden_mag": hidden_mag, "label_mag": label_mag}


    def sample_conditioned(
        self,
        gibbs_steps: int,
        targets: torch.Tensor | np.ndarray,
        visible: torch.Tensor | None = None,
        beta: float = 1.0,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Samples from the annaRBM conditioned on the target labels. During the sampling, the labels are kept fixed
        and the visible and hidden units are sampled alternatively. The visible's conditional probability distribution
        is returned.

        Args:
            gibbs_steps (int): Number of Alternate Gibbs Sampling steps.
            targets (torch.Tensor | np.ndarray): Target labels.
            visible (torch.Tensor | None, optional): Visible units. If None, it will be initialized randomly. Defaults to None.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            Dict[str, torch.Tensor]: Sampled visible and hidden units.
        """
        if isinstance(targets, np.ndarray):
            targets = torch.tensor(targets, device=self.device, dtype=self.dtype)
        elif isinstance(targets, torch.Tensor):
            targets = targets.to(self.device, dtype=self.dtype)
        num_samples = targets.shape[0]
        if visible is None:
            visible = self.init_chains(num_samples=num_samples)["visible"]
        else:
            if visible.shape[0] != num_samples:
                raise ValueError(f"The number of visible units ({visible.shape[0]}) and targets ({num_samples}) must be the same.")

        visible, hidden, visible_mag, hidden_mag = _sample_conditioned(
            gibbs_steps=gibbs_steps,
            label=targets,
            visible=visible,
            params=self.params,
            beta=beta,
        )

        return {"visible": visible, "hidden": hidden, "visible_mag": visible_mag, "hidden_mag": hidden_mag}


    def predict_labels(
        self,
        visible: torch.Tensor | np.ndarray,
        beta: float = 1.0,
        **kwargs,
    ) -> torch.Tensor:
        """Returns the probability distribution of each label given the visible units: p(l|v).

        Args:
            visible (torch.Tensor | np.ndarray): Visible units.
            beta (float, optional): Inverse temperature. Defaults to 1.0.

        Returns:
            torch.Tensor: Labels's probability distribution: p(l|v).
        """
        if isinstance(visible, np.ndarray):
            visible = torch.tensor(visible, device=self.device, dtype=self.dtype)
        
        p_labels = _predict_labels(
            visible=visible,
            params=self.params,
            beta=beta,            
        )
        
        return p_labels
    
    
    def get_pattern(self, label_idx: int) -> torch.Tensor:
        """Returns the visible pattern associated to the given label index.

        Args:
            label_idx (int): Index of the label.

        Returns:
            torch.Tensor: Visible pattern of dimension (num_visibles,) associated to the label.
        """
        mag_prototype = self.params["label_matrix"][label_idx]
        pattern = (self.params["weight_matrix"] @ mag_prototype)
        
        return pattern
    
    
    def update_weights_AIS(
        self,
        prev_model: annaRBM,
        chains: Dict[str, torch.Tensor],
        log_weights: torch.Tensor,
    ) -> torch.Tensor:
        """Update the weights used during the trajectory Annealed Importance Sampling (AIS) algorithm.

        Args:
            prev_model (annRBM): Model at time t-1.
            chains (torch.Tensor): Chains at time t-1.
            log_weights (torch.Tensor): Log-weights at time t-1.

        Returns:
            torch.Tensor: Log-weights at time t.
        """
        if prev_model.params is None:
            raise ValueError("Previous model parameters are not initialized.")
        return _update_weights_AIS(prev_model.params, self.params, chains, log_weights)
    
    
    def compute_log_likelihood(
        self,
        visible: torch.Tensor,
        weight: torch.Tensor,
        logZ: float,
        **kwargs,
    ) -> float:
        """Computes the log-likelihood of the model on the given configuration.

        Args:
            visible (torch.Tensor): Visible units.
            weight (torch.Tensor): weights of the sequences.
            logZ (float): Log partition function.

        Returns:
            float: Log-likelihood of the model on the given configuration.
        """
        return _compute_log_likelihood(visible, weight, self.params, logZ)
    
    
    def init_chains(
        self,
        num_samples: int,
        use_profile: bool = False,
    ) -> Dict[str, torch.Tensor]:
        """Initialize the Markov chains for the RBM by sampling a uniform distribution on the visible layer and the labels
        and sampling the hidden layer according to the visible one. If use_profile is True, the visible units and the label
        are sampled from the profile model using the local fields.

        Args:
            num_samples (int): Number of parallel chains.
            use_profile (bool, optional): Whether to use the profile model. Defaults to False.
    
        Returns:
            Dict[str, torch.Tensor]: Initial Markov chain.
        """
        return _init_chains(num_samples, self.params, use_profile)
    
    
    @staticmethod
    def save_chains(
        filename: str,
        visible: torch.Tensor,
        label: torch.Tensor,
        **kwargs,
    ) -> None:
        """Save the persistent chains on a fasta file.

        Args:
            filename (str): Path to the fasta file.
            visible (torch.Tensor): Visible units of the chains.
            label (torch.Tensor): Labels of the chains.
        """
        _save_chains(
            filename=filename,
            visible=visible,
            label=label,
            alphabet="01",
        )
        
    
    def load_chains(
        self,
        filename: str,
        device: torch.device,
        dtype: torch.dtype,
        **kwargs,
    ) -> Dict[str, torch.Tensor]:
        """Load the persistent chains from a fasta file.

        Args:
            filename (str): Path to the fasta file.
            device (torch.device): Device on which to load the chains.
            dtype (torch.dtype): Data type for the chains.

        Returns:
            Dict[str, torch.Tensor]: Visible and hidden units and labels of the chains.
        """
        headers, sequences = import_from_fasta(filename)
        
        label = np.vectorize(lambda x: np.array([int(i) for i in x]), signature="() -> (l)")(headers)
        visible = np.vectorize(lambda x: np.array([int(i) for i in x]), signature="() -> (l)")(sequences)
        label = torch.tensor(label, device=device, dtype=dtype)
        visible = torch.tensor(visible, device=device, dtype=dtype)
        if self.params is not None:
            hidden = self.sample_hiddens(visible, label)["hidden"]
        else:
            hidden = torch.zeros((visible.shape[0],), device=device, dtype=dtype)

        return {"visible": visible, "hidden": hidden, "label": label}
    
    
    def compute_gradient(
        self,
        data: Dict[str, torch.Tensor],
        chains: Dict[str, torch.Tensor],
        pseudo_count: float = 0.0,
        centered: bool = True,
        lambda_l1: float = 0.0,
        lambda_l2: float = 0.0,
        eta: float = 1.0,
    ) -> None:
        """Computes the gradient of the log-likelihood and stores it.

        Args:
            data (Dict[str, torch.Tensor]): Data batch.
            chains (Dict[str, torch.Tensor]): Chains.
            pseudo_count (float, optional): Pseudo count to be added to the data frequencies. Defaults to 0.0.
            centered (bool, optional): Centered gradient. Defaults to True.
            lambda_l1 (float, optional): L1 regularization weight. Defaults to 0.0.
            lambda_l2 (float, optional): L2 regularization weight. Defaults to 0.0.
            eta (float, optional): Relative contribution of the label term. Defaults to 1.0.
        """
        _compute_gradient(
            data=data,
            chains=chains,
            params=self.params,
            pseudo_count=pseudo_count,
            centered=centered,
            lambda_l1=lambda_l1,
            lambda_l2=lambda_l2,
            eta=eta,
        )
        
    
    def zerosum_gauge(self) -> None:
        """Uneffective method for the binary annaRBM."""
        # The zerosum gauge is not applicable to the binary RBM, so we do nothing here.
        pass
        
    
    def init_parameters(
        self,
        num_visibles: int,
        num_hiddens: int,
        num_labels: int,
        frequencies_visibles: torch.Tensor | None = None,
        frequencies_labels: torch.Tensor | None = None,
        std_init: float = 1e-4,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
        **kwargs,
    ) -> None:
        """
        Initialize the parameters of the model. Hidden biases are set to 0,
        visible and label biases are set to the independent-site model value using the frequencies of the dataset (if provided),
        and the weight and label matrices are initialized with a Gaussian distribution of sigma std_init.

        Args:
            num_visibles (int): Number of visible units.
            num_hiddens (int): Number of hidden units.
            num_labels (int): Number of labels.
            frequencies_visibles (torch.Tensor): Empirical frequencies of the visible units. Defaults to None.
            frequencies_labels (torch.Tensor): Empirical frequencies of the labels. Defaults to None.
            std_init (float, optional): Standard deviation of the weight matrix. Defaults to 1e-4.
            device (torch.device): Device for the parameters.
            dtype (torch.dtype): Data type for the parameters.
        """
        self.device = device
        self.dtype = dtype
        self.params = _init_parameters(
            num_visibles=num_visibles,
            num_hiddens=num_hiddens,
            num_labels=num_labels,
            frequencies_visibles=frequencies_visibles,
            frequencies_labels=frequencies_labels,
            std_init=std_init,
            device=self.device,
            dtype=self.dtype
        )
        
    
    def num_visibles(self) -> int:
        """Returns the number of visible units.

        Returns:
            int: Number of visible units.
        """
        return super().num_visibles()
    
    
    def num_hiddens(self) -> int:
        """Returns the number of hidden units.

        Returns:
            int: Number of hidden units.
        """
        return super().num_hiddens()
    
    
    def num_classes(self) -> int:
        """Returns the number of label classes.

        Returns:
            int: Number of label classes.
        """
        return super().num_classes()
    
    
    def logZ0(self) -> float:
        """Computes the initial log partition function for the annaRBM.

        Returns:
            float: Initial log partition function.
        """
        logZ_visibles = torch.log(1.0 + torch.exp(self.params["vbias"])).sum().item()
        logZ_labels = torch.log(1.0 + torch.exp(self.params["lbias"])).sum().item()
        logZ_hiddens = torch.log(1.0 + torch.exp(self.params["hbias"])).sum().item()
        return logZ_visibles + logZ_labels + logZ_hiddens
        


from typing import Any
import numpy as np
import pandas as pd

from torch.utils.data import Dataset
import torch
from torch.nn.functional import one_hot
from adabmDCA.fasta import get_tokens, compute_weights, import_from_fasta, encode_sequence

class annaDataset(Dataset):
    
    def __init__(
        self,
        path_data: str,
        path_ann: str | None = None,
        column_names: str = "name",
        column_sequences: str = "sequence",
        column_labels: str = "label",
        is_binary: bool = False,
        alphabet: str = "protein",
        clustering_th: float = 0.8,
        no_reweighting: bool = True,
        path_weights: str | None = None,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
    ):
        self.device = device
        self.dtype = dtype
        self.data = pd.DataFrame()
        self.data_one_hot = torch.tensor([], dtype=dtype, device=device)
        self.labels_one_hot = torch.tensor([], dtype=dtype, device=device)
        self.weights = torch.tensor([], dtype=dtype, device=device)
        self.is_binary = is_binary
        self.column_names = column_names
        self.column_sequences = column_sequences
        self.column_labels = column_labels
        self.L = 0
        self.q = 0

        # check if path_data points to a text file
        if path_data.endswith(".txt") or path_data.endswith(".dat"):
            self.parse_text(path_data)
            if path_ann is not None:
                self.parse_csv(
                    path_ann=path_ann,
                    column_names=column_names,
                    column_sequences=column_sequences,
                    column_labels=column_labels,
                )
            else:
                raise ValueError("The annotations file must be provided if data are taken from a text file.")
        elif path_data.endswith(".csv"):
            self.parse_csv(
                path_ann=path_data,
                column_names=column_names,
                column_sequences=column_sequences,
                column_labels=column_labels,
            )
        else:
            # check that the input file is a valid fasta file
            with open(path_data, "r") as f:
                first_line = f.readline()
                if first_line.startswith(">"):
                    self.parse_fasta(path_data)
                else:
                    raise ValueError("Input data not recognized as a text file nor a .csv file. Invalid FASTA format.")
            if path_ann is not None:
                self.parse_csv(
                    path_ann=path_ann,
                    column_names=column_names,
                    column_sequences=column_sequences,
                    column_labels=column_labels,
                )
            else:
                raise ValueError("The annotations file must be provided if data are taken from a fasta file.")

        if is_binary:
            self.tokens = "01"
        else:
            self.tokens = get_tokens(alphabet)

        # encode sequences
        self.data_one_hot = torch.tensor(
                encode_sequence(list(self.data[column_sequences].values), self.tokens),
                dtype=self.dtype,
                device=self.device,
            )
        # one-hot representation of sequences and labels
        self.parse_labels()
        
        self.L = self.data_one_hot.shape[1]
        self.q = len(self.tokens)
            
        if no_reweighting:
            self.weights = torch.ones(len(self.data), 1, dtype=dtype, device=device).view(-1, 1)
        else:
            if path_weights is not None:
                self.weights = torch.tensor(
                    np.loadtxt(path_weights),
                    dtype=dtype,
                    device=device,
                ).view(-1, 1)
            else:
                # Here, data_one_hot is still a (M, L) tensor
                self.weights = compute_weights(
                    data=self.data_one_hot,
                    th=clustering_th,
                    device=device,
                    dtype=dtype,
                ).view(-1, 1)
                
        if not is_binary:
            self.data_one_hot = one_hot(self.data_one_hot.long(), num_classes=self.q).to(self.dtype)


    def parse_labels(self):
        labels = self.data[self.column_labels]
        unique_labels = np.unique(labels.dropna())
        self.idx_to_label = {i: label for i, label in enumerate(unique_labels)}
        self.label_to_idx = {label: i for i, label in self.idx_to_label.items()}
        label_indices = labels.map(self.label_to_idx).fillna(-1).astype(int)
        num_classes = len(unique_labels)
        one_hot = np.zeros((len(labels), num_classes), dtype=np.float32)
        valid = label_indices != -1
        one_hot[valid, label_indices[valid]] = 1.0
        self.labels_one_hot = torch.tensor(
            one_hot,
            dtype=self.dtype,
            device=self.device,
        )
        
        
    def parse_csv(
        self,
        path_ann: str,
        column_names: str = "name",
        column_sequences: str = "sequence",
        column_labels: str = "label",
    ):
        
        df = pd.read_csv(path_ann, na_values=[""])
        assert column_names.lower() in df.columns, f"Column {column_names} not found in the labels file."
        assert column_labels.lower() in df.columns, f"Column {column_labels} not found in the labels file."
        
        # If all data comes from the csv file:
        if len(self.data) == 0:
            assert column_sequences.lower() in df.columns, f"Column {column_sequences} not found in the labels file."
            self.data = df.filter(items=[column_names, column_sequences, column_labels])
        # Else, names and data have been already parsed from a fasta file:
        else:
            # merge df with self.data on the names
            self.data = pd.merge(
                self.data,
                df.filter(items=[column_names, column_labels]),
                on=column_names,
                how="left",
            )
        
            
    def parse_fasta(
        self,
        path_data: str,
    ):
        names, sequences = import_from_fasta(path_data)
        self.data = pd.DataFrame.from_dict(
            {"name": names, "sequence": sequences}  
        )


    def parse_text(
        self,
        path_data: str,
    ):
        data = torch.tensor(
            np.loadtxt(path_data),
            dtype=self.dtype,
            device=self.device,
        )
        names = np.arange(len(data)).astype(str)
        self.data = pd.DataFrame.from_dict(
            {"name": names, "sequence": data}  
        )
    
    def to_label(
        self,
        labels: torch.Tensor | np.ndarray | list,
    ) -> np.ndarray:
        """Converts the one-hot encoded labels (or their probability) to the original labels.
        
        Args:
            labels (torch.Tensor | np.ndarray | list): One-hot encoded labels or their probability.
            
        Returns:
            torch.Tensor: Original labels.
        """
        if isinstance(labels, torch.Tensor):
            labels = labels.cpu().numpy()
        elif isinstance(labels, list):
            labels = np.array(labels)
        if len(labels.shape) == 1:
            labels = np.expand_dims(labels, axis=0)
            
        return np.array([self.idx_to_label[np.argmax(l).item()] for l in labels])
    
    
    def to_one_hot(
        self,
        labels: torch.Tensor | np.ndarray | list,
    ) -> torch.Tensor:
        """Converts the original labels to one-hot encoded labels.
        
        Args:
            labels (torch.Tensor | np.ndarray | list): Original labels.
            
        Returns:
            torch.Tensor: One-hot encoded labels.
        """
        if isinstance(labels, torch.Tensor):
            labels = labels.cpu().numpy().tolist()
        elif isinstance(labels, np.ndarray):
            labels = labels.tolist()
        
        one_hot_labels = np.zeros((len(labels), len(self.idx_to_label)), dtype=np.float32)
        for i, label in enumerate(labels):
            if label in self.label_to_idx:
                one_hot_labels[i, self.label_to_idx[label]] = 1.0
        return torch.tensor(one_hot_labels, dtype=self.dtype, device=self.device)


    def __len__(self):
        return len(self.data)


    def __getitem__(self, idx: int) -> Any:
        return {
            "visible": self.data_one_hot[idx],
            "label": self.labels_one_hot[idx],
            "weight": self.weights[idx],
        }
    
    
    def get_num_residues(self) -> int:
        """Returns the number of residues (L) in the multi-sequence alignment.

        Returns:
            int: Length of the MSA.
        """
        return self.L
    
    
    def get_num_states(self) -> int:
        """Returns the number of states (q) in the alphabet.

        Returns:
            int: Number of states.
        """
        return self.q
    
    
    def get_num_classes(self) -> int:
        """Returns the number of categories in the labels.

        Returns:
            int: Number of categories.
        """
        return self.labels_one_hot.shape[1]
    
    
    def get_effective_size(self) -> int:
        """Returns the effective size (Meff) of the dataset.

        Returns:
            int: Effective size of the dataset.
        """
        return int(self.weights.sum())
    
    
    def shuffle(self) -> None:
        """Shuffles the dataset.
        """
        perm = torch.randperm(len(self.data))
        # shuffle the data, one-hot encoded sequences, names, weights, and labels
        self.data = self.data.iloc[perm.numpy()].reset_index(drop=True)
        self.data_one_hot = self.data_one_hot[perm]
        self.weights = self.weights[perm]
        self.labels_one_hot = self.labels_one_hot[perm]


import torch

@torch.jit.script
def _get_freq_single_point(
    data: torch.Tensor,
    weights: torch.Tensor | None,
    pseudo_count: float = 0.0,
) -> torch.Tensor:    
    M = len(data)
    if weights is not None:
        norm_weights = weights.reshape(M, 1) / weights.sum()
    else:
        norm_weights = torch.ones((M, 1), device=data.device) / M

    frequencies = (data * norm_weights).sum(dim=0)

    return (1. - pseudo_count) * frequencies + (pseudo_count / 2)


def get_freq_single_point(
    data: torch.Tensor,
    weights: torch.Tensor | None,
    pseudo_count: float = 0.,
) -> torch.Tensor:
    """Computes the single point frequencies of the input binary data.
    Args:
        data (torch.Tensor): Binary data array.
        weights (torch.Tensor | None): Weights of the sequences.
        pseudo_count (float, optional): Pseudo count to be added to the frequencies. Defaults to 0.0.
    
    Raises:
        ValueError: If the input data is not a 2D tensor.

    Returns:
        torch.Tensor: Single point frequencies.
    """
    if data.dim() != 2:
        raise ValueError(f"Expected data to be a 2D tensor, but got {data.dim()}D tensor instead")
    
    return _get_freq_single_point(data, weights, pseudo_count)


@torch.jit.script
def _get_freq_two_points(
    data: torch.Tensor,
    weights: torch.Tensor | None,
    pseudo_count: float=0.0,
) -> torch.Tensor:
    
    M, L = data.shape
    
    if weights is not None:
        norm_weights = weights.reshape(M, 1) / weights.sum()
    else:
        norm_weights = torch.ones((M, 1), device=data.device) / M
    
    fij = (data * norm_weights).T @ data
    # Apply the pseudo count
    fij = (1. - pseudo_count) * fij + (pseudo_count / 4)
    # Diagonal terms must represent the single point frequencies
    fi = get_freq_single_point(data, weights, pseudo_count).ravel()
    # Apply the pseudo count on the single point frequencies
    fij_diag = (1. - pseudo_count) * fi + (pseudo_count / 2)
    # Set the diagonal terms of fij to the single point frequencies
    fij = torch.diagonal_scatter(fij, fij_diag, dim1=0, dim2=1)
    
    return fij


def get_freq_two_points(
    data: torch.Tensor,
    weights: torch.Tensor | None,
    pseudo_count: float=0.0,
) -> torch.Tensor:
    """
    Computes the 2-points statistics of the input binary matrix

    Args:
        data (torch.Tensor): Binary data array.
        weights (torch.Tensor | None): Array of weights to assign to the sequences of shape.
        pseudo_count (float, optional): Pseudo count for the single and two points statistics. Acts as a regularization. Defaults to 0.0.
    
    Raises:
        ValueError: If the input data is not a 2D tensor.

    Returns:
        torch.Tensor: Matrix of two-point frequencies.
    """
    if data.dim() != 2:
        raise ValueError(f"Expected data to be a 2D tensor, but got {data.dim()}D tensor instead")
    
    return _get_freq_two_points(data, weights, pseudo_count)


from typing import Dict
import torch

from adabmDCA.stats import get_freq_single_point as get_freq_single_point_cat # for categorical data
from annadca.rbm.binary.stats import get_freq_single_point as get_freq_single_point_bin # for binary data


@torch.jit.script
def _compute_gradient(
    data: Dict[str, torch.Tensor],
    chains: Dict[str, torch.Tensor],
    params: Dict[str, torch.Tensor],
    pseudo_count: float = 0.0,
    centered: bool = True,
    lambda_l1: float = 0.0,
    lambda_l2: float = 0.0,
    eta: float = 1.0,
) -> None:
    """Computes the gradient of the Log-Likelihood for the binary annaRBM.

    Args:
        data (Dict[str, torch.Tensor]): Data batch.
        chains (Dict[str, torch.Tensor]): Chains.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        pseudo_count (float, optional): Pseudo count to be added to the data frequencies. Defaults to 0.0.
        centered (bool, optional): Whether to use centered gradients. Defaults to True.
        lambda_l1 (float, optional): L1 regularization weight. Defaults to 0.0.
        lambda_l2 (float, optional): L2 regularization weight. Defaults to 0.0.
        eta (float, optional): Relative contribution of the label term. Defaults to 1.0.
    """
    # Normalize the weights
    data["weight"] = data["weight"] / data["weight"].sum()
    nchains = len(chains["visible"])
    
    # Averages over data and generated samples
    v_data_mean = get_freq_single_point_cat(data["visible"], weights=data["weight"], pseudo_count=pseudo_count)
    h_data_mean = get_freq_single_point_bin(data["hidden"], weights=data["weight"], pseudo_count=pseudo_count)
    l_data_mean = get_freq_single_point_bin(data["label"], weights=data["weight"], pseudo_count=pseudo_count)
    v_gen_mean = get_freq_single_point_cat(chains["visible"], weights=None, pseudo_count=0.0)
    h_gen_mean = chains["hidden"].mean(0)
    l_gen_mean = chains["label"].mean(0)
    
    if centered:
        # Centered variables
        v_data_centered = data["visible"] - v_data_mean
        h_data_centered = data["hidden"] - h_data_mean
        v_gen_centered = chains["visible"] - v_data_mean
        h_gen_centered = chains["hidden"] - h_data_mean
        l_data_centered = data["label"] - l_data_mean
        l_gen_centered = chains["label"] - l_data_mean
        
        # Gradient
        grad_weight_matrix = (
            torch.einsum("nlq, np -> lqp", v_data_centered * data["weight"].view(-1, 1, 1), h_data_centered) -
            torch.einsum("nlq, np -> lqp", v_gen_centered, h_gen_centered) / nchains
        )
        grad_label_matrix = (
            (l_data_centered * data["weight"]).T @ h_data_centered -
            (l_gen_centered.T @ h_gen_centered) / nchains
        )
        grad_vbias = v_data_mean - v_gen_mean - (grad_weight_matrix @ h_data_mean)
        grad_hbias = h_data_mean - h_gen_mean - torch.einsum("lq, lqp -> p", v_data_mean, grad_weight_matrix)
        grad_lbias = l_data_mean - l_gen_mean - (grad_label_matrix @ h_data_mean)
        
        # regularization
        grad_weight_matrix -= lambda_l1 * params["weight_matrix"].sign() + lambda_l2 * params["weight_matrix"]
        #grad_label_matrix -= lambda_l1 * params["label_matrix"].sign() + lambda_l2 * params["label_matrix"]
        
    else:
        # Gradient
        grad_weight_matrix = (
            torch.einsum("nlq, np -> lqp", data["visible"] * data["weight"].view(-1, 1, 1), data["hidden"]) -
            torch.einsum("nlq, np -> lqp", chains["visible"], chains["hidden"]) / nchains
        )
        grad_label_matrix = (
            (data["label"] * data["weight"]).T @ data["hidden"] -
            (chains["label"].T @ chains["hidden"]) / nchains
        )
        grad_vbias = v_data_mean - v_gen_mean
        grad_hbias = h_data_mean - h_gen_mean
        grad_lbias = l_data_mean - l_gen_mean
        
        # regularization
        grad_weight_matrix -= lambda_l1 * params["weight_matrix"].sign() + lambda_l2 * params["weight_matrix"]
        #grad_label_matrix -= lambda_l1 * params["label_matrix"].sign() + lambda_l2 * params["label_matrix"]

    # Attach the gradients to the parameters
    params["weight_matrix"].grad.set_(grad_weight_matrix)
    params["label_matrix"].grad.set_(eta * grad_label_matrix)
    params["vbias"].grad.set_(grad_vbias)
    params["hbias"].grad.set_(grad_hbias)
    params["lbias"].grad.set_(eta * grad_lbias)
        


from typing import Dict

import numpy as np
import torch
from torch.nn.functional import one_hot


def _init_parameters(
    num_visibles: int,
    num_hiddens: int,
    num_labels: int,
    frequencies_visibles: torch.Tensor | None = None,
    frequencies_labels: torch.Tensor | None = None,
    std_init: float = 1e-4,
    device: torch.device = torch.device("cpu"),
    dtype: torch.dtype = torch.float32,
) -> Dict[str, torch.Tensor]:
    """
    Hidden biases are set to 0, visible and label biases are set to the independent-site model value using the frequencies of the dataset,
    and the weight and label matrices are initialized with a Gaussian distribution of sigma std_init.
    
    Args:
        num_visibles (int): Number of visible units.
        num_hiddens (int): Number of hidden units.
        num_labels (int): Number of labels.
        frequencies_visibles (torch.Tensor): Empirical frequencies of the visible units. Defaults to None.
        frequencies_labels (torch.Tensor): Empirical frequencies of the labels. Defaults to None.
        std_init (float, optional): Standard deviation of the weight matrix. Defaults to 1e-4.
        device (torch.device): Device for the parameters.
        dtype (torch.dtype): Data type for the parameters.
    
    Returns:
        Dict[str, torch.Tensor]: Parameters of the model.
    """
    if isinstance(frequencies_visibles, np.ndarray):
        frequencies_visibles = torch.from_numpy(frequencies_visibles).to(device=device, dtype=dtype)
    if isinstance(frequencies_labels, np.ndarray):
        frequencies_labels = torch.from_numpy(frequencies_labels).to(device=device, dtype=dtype)
        
    weight_matrix = torch.randn(size=(num_visibles, num_hiddens), device=device, dtype=dtype) * std_init
    label_matrix = torch.randn(size=(num_labels, num_hiddens), device=device, dtype=dtype) * std_init
    hbias = torch.zeros(num_hiddens, device=device, dtype=dtype)
    
    if frequencies_visibles is None:
        vbias = torch.zeros(num_visibles, device=device, dtype=dtype)
    else:
        vbias = (torch.log(frequencies_visibles) - torch.log(1.0 - frequencies_visibles)).to(device)
    
    if frequencies_labels is None:
        lbias = torch.zeros(num_labels, device=device, dtype=dtype)
    else:
        lbias = (torch.log(frequencies_labels) - torch.log(1.0 - frequencies_labels)).to(device)
    
    return {"vbias": vbias, "hbias": hbias, "lbias": lbias, "weight_matrix": weight_matrix, "label_matrix": label_matrix}


def _init_chains(
    num_samples: int,
    params: Dict[str, torch.Tensor],
    use_profile: bool = False,
) -> Dict[str, torch.Tensor]:
    """Initialize the Markov chains for the RBM by sampling a uniform distribution on the visible layer and the labels
    and sampling the hidden layer according to the visible one. If use_profile is True, the visible units and the label
    are sampled from the profile model using the local fields.

    Args:
        num_samples (int): Number of parallel chains.
        params (Dict[str, torch.Tensor]): Parameters of the model.
        use_profile (bool, optional): Whether to use the profile model. Defaults to False.

    Returns:
        Dict[str, torch.Tensor]: Initial Markov chain.
    """
    num_visibles = len(params["vbias"])
    num_labels = len(params["lbias"])
    
    if use_profile:
        mv = torch.sigmoid(params["vbias"]).repeat(num_samples, 1)
        ml = torch.sigmoid(params["lbias"]).repeat(num_samples, 1)
        
    else:
        mv = torch.ones(
            size=(num_samples, num_visibles),
            device=params["vbias"].device,
            dtype=params["vbias"].dtype) / 2
        ml = torch.ones(
            size=(num_samples, num_labels),
            device=params["lbias"].device,
            dtype=params["lbias"].dtype) / num_labels
        
    visible = torch.bernoulli(mv)
    # label is extracted from a multinomial
    label = one_hot(torch.multinomial(ml, 1), num_classes=num_labels).to(params["weight_matrix"].dtype).squeeze(1)

    mh = torch.sigmoid((params["hbias"] + (visible @ params["weight_matrix"]) + (label @ params["label_matrix"])))
    hidden = torch.bernoulli(mh)
    
    return {"visible": visible, "hidden": hidden, "label": label}



import subprocess
import sys
import os

def main():
    # Get the directory of the current script
    SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))

    # Check if the first positional argument is provided
    if len(sys.argv) < 2:
        print("Error: No command provided. Use 'train'.")
        sys.exit(1)

    # Assign the first positional argument to a variable
    COMMAND = sys.argv[1]

    # Map the command to the corresponding script
    match COMMAND:
        case "train":
            SCRIPT = "train.py"
        case _:
            print(f"Error: Invalid command '{COMMAND}'. Use 'train'.")
            sys.exit(1)

    # Run the corresponding Python script with the remaining optional arguments
    REPO_SCRIPTS = os.path.join(SCRIPT_DIR, "scripts")
    script_path = os.path.join(REPO_SCRIPTS, SCRIPT)
    proc = subprocess.call(
        [sys.executable, script_path] + sys.argv[2:],
    )

if __name__ == "__main__":
    main()


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec


def _plot_hist(ax, data1, data2, color, dim, orientation='vertical'):    
    ax.hist(data1[:, dim], bins=30, color=color, histtype='step', zorder=0, density=True, orientation=orientation, lw=1)
    ax.hist(data2[:, dim], bins=30, color=color, histtype='bar', zorder=1, density=True, orientation=orientation, alpha=0.3)
    ax.axis('off')
    

def plot_PCA(
    fig: plt.figure,
    data1: np.array,
    labels1: np.ndarray,
    data2: np.array,
    labels2: np.ndarray,
    label_to_color: dict,
    dim1: int,
    dim2: int,
    title: str,
    split_labels: bool = False,
) -> plt.figure:
    """Plot the scatter plot of data1 and data2 in the space defined by the (dim1, dim2) principal components.
    Also plots the histograms projected along the x and y axes.
    
    Args:
        fig (plt.figure): Matplotlib figure.
        data1 (np.array): Data array.
        labels1 (np.ndarray): Labels array.
        data2 (np.array): Data array.
        labels2 (np.ndarray): Labels array.
        label_to_color (dict): Dictionary of label to color.
        dim1 (int): First principal component.
        dim2 (int): Second principal component.
        title (str): Title of the plot.
        split_labels (bool, optional): If True, the histograms are split by label. Defaults to False.
    """
    gs = GridSpec(4, 4)

    ax_scatter = fig.add_subplot(gs[1:4, 0:3])
    ax_hist_x = fig.add_subplot(gs[0, 0:3])
    ax_hist_y = fig.add_subplot(gs[1:4, 3])
    
    unique_labels = np.unique(labels1)
    for label in unique_labels:
        mask1 = labels1 == label
        mask2 = labels2 == label
        ax_scatter.scatter(data1[mask1, dim1], data1[mask1, dim2], color=label_to_color[label], s=50, zorder=0, alpha=0.3, label=label)
        if split_labels:
            _plot_hist(ax_hist_x, data1[mask1], data2[mask2], color=label_to_color[label], dim=dim1, orientation='vertical')
            _plot_hist(ax_hist_y,data1[mask1], data2[mask2], color=label_to_color[label], dim=dim2, orientation='horizontal')
    
    for label in unique_labels:
        mask = labels2 == label
        ax_scatter.scatter(data2[mask, dim1], data2[mask, dim2], color=label_to_color[label], s=20, zorder=2, edgecolor='black', marker='o', alpha=1, linewidth=0.4)
    
    if not split_labels:
        _plot_hist(ax_hist_x, data1, data2, 'red', dim1)
        _plot_hist(ax_hist_y, data1, data2, 'red', dim2, orientation='horizontal')
    
    ax_scatter.set_xlabel(f"PC {dim1 + 1}")
    ax_scatter.set_ylabel(f"PC {dim2 + 1}")
    
    fig.suptitle(title)
    return fig


def plot_mixing_time(results: dict) -> plt.figure:
    """Plot the mixing time of the model.
    
    Args:
        results (dict): Dictionary containing the results of the resampling.
        
    Returns:
        plt.figure: Matplotlib figure.
    """
    fig, ax = plt.subplots(figsize=(6, 4), dpi=96)
    ax.plot(results["t_half"], results["seqid_t"], label=r"SeqID$(t)$", color="#002CFF")  # Blue
    ax.fill_between(results["t_half"],
                     np.array(results["seqid_t"]) - np.array(results["std_seqid_t"]),
                     np.array(results["seqid_t"]) + np.array(results["std_seqid_t"]),
                     color="#002CFF", alpha=0.2)
    ax.plot(results["t_half"], results["seqid_t_t_half"], label=r"SeqID$(t, t/2)$", color="#67BAA6")  # Orange
    ax.fill_between(results["t_half"],
                     np.array(results["seqid_t_t_half"]) - np.array(results["std_seqid_t_t_half"]),
                     np.array(results["seqid_t_t_half"]) + np.array(results["std_seqid_t_t_half"]),
                     color="#67BAA6", alpha=0.2)
    ax.set_xlabel(r"$t/2$ (Sweeps)")
    ax.set_ylabel("Sequence Identity")
    ax.legend(loc='upper right')
    ax.set_title("Mixing time")


    # Add annotation for mixing time
    ax.annotate(r"$t^{\mathrm{mix}}=$" + f"{results['t_half'][-1]}", xy=(0.96, 0.7), xycoords='axes fraction', fontsize=15,
                 verticalalignment='top', horizontalalignment='right', bbox=dict(facecolor='white', alpha=0.5))

    return fig


from setuptools import setup, find_packages

setup(
    name='annadca',
    version='0.2.1',
    author='Lorenzo Rosset, Aurélien Decelle, Beatriz Seoane, Francesco Zamponi, Martin Weigt',
    maintainer='Lorenzo Rosset',
    author_email='rosset.lorenzo@gmail.com',
    description='Annotation Assisted Direct Coupling Analysis',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/rossetl/annaDCA',
    packages=find_packages(include=['annadca', 'annadca.*']),
    include_package_data=True,
    python_requires='>=3.10',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'annadca=annadca.cli:main',
        ],
    },
    install_requires=[
        'adabmDCA>=0.7.0'
    ],
)
